// LuaFilter.cpp : Implementation of CLuaFilter

#include "stdafx.h"
#include "LuaFilter.h"
#include <stdidcmd.h>
#include <set>

class StringSorter
{
public:
	bool operator()( std::wstring const& left, std::wstring const& right ) const
	{
		// if different when case-insensitive then sort
		int cmp = _wcsicmp( left.c_str(), right.c_str() );
		if( cmp != 0 )
			return cmp < 0;

		// otherwise sort lowercase before uppercase
		return wcscmp( left.c_str(), right.c_str() ) > 0;
   	}
};

GUID const CMDSETID_StandardCommandSet2K = {0x1496A755,0x94DE,0x11D0,{0x8C,0x3F,0x00,0xC0,0x4F,0xC2,0xAA,0xE2}};

// CLuaFilter

HRESULT CLuaFilter::Attach( CComPtr<IVsTextView> const& view )
{
	// cache the view
	m_view = view;

	// attach ourselves
	return m_view->AddCommandFilter( this, &m_next );
}

void CLuaFilter::Detach()
{
	// detach ourselves
	m_view->RemoveCommandFilter( this );
	m_view.Release();
}

STDMETHODIMP CLuaFilter::QueryStatus( GUID const* commandGroup, 
									  ULONG commandCount, 
									  OLECMD commands[], 
									  OLECMDTEXT* commandText)
{
	// check arguments
	if( !commandGroup || !commands )
		return E_INVALIDARG;

	// if we're asked to only autocomplete say that we can
	if( IsEqualGUID( CMDSETID_StandardCommandSet2K, *commandGroup ) && commandCount == 1 )
    {
		if( commands[0].cmdID == ECMD_COMPLETEWORD )
		{
			commands[0].cmdf = OLECMDF_SUPPORTED | OLECMDF_ENABLED;
			return S_OK;
		}
    }

	// pass on
	return m_next->QueryStatus( commandGroup, commandCount, commands, commandText );
}

STDMETHODIMP CLuaFilter::Exec( GUID const* commandGroup, 
							   DWORD commandID, 
							   DWORD options, 
							   VARIANT* varIn, 
							   VARIANT* varOut )
{
	// check arguments
	if( !commandGroup )
		return E_INVALIDARG;

	// process auto-complete requests
	if( IsEqualGUID( CMDSETID_StandardCommandSet2K, *commandGroup ) )
	{
		switch( commandID )
		{
		case ECMD_COMPLETEWORD:
			// build a new list of idents
			ResetList();

			// show the list
			m_view->UpdateCompletionStatus( this, UCS_NAMESCHANGED | UCS_COMPLETEWORD );
			return S_OK;

        case ECMD_TYPECHAR:
			// check for '.' or ':'
            if( varIn->vt == VT_UI2 && ( varIn->uiVal == '.' || varIn->uiVal == ':' ) )
			{
				// pass on the key
				m_next->Exec( commandGroup, commandID, options, varIn, varOut );

				// build a new list of idents
				ResetList();

				// show the list
				m_view->UpdateCompletionStatus( this, UCS_NAMESCHANGED );
				return S_OK;
			}
		}
	}

	// pass on
	return m_next->Exec( commandGroup, commandID, options, varIn, varOut );
}

STDMETHODIMP_( DWORD ) CLuaFilter::GetFlags()
{
	// no flags
	return 0;
}

STDMETHODIMP_( long ) CLuaFilter::GetCount()
{
	return long( m_identifiers.size() );
}

STDMETHODIMP CLuaFilter::GetDisplayText( long index, WCHAR const** text, long* glyph )
{
	if( index < 0 || index >= long( m_identifiers.size() ) )
		return E_INVALIDARG;
	if( !text )
		return E_INVALIDARG;

	// store the text
	*text = m_identifiers[index].c_str();

	// store the image
	if( glyph )
		*glyph = 0;

	// done
	return S_OK;
}

STDMETHODIMP CLuaFilter::GetImageList( HANDLE* /*images*/ )
{
	return E_NOTIMPL;
}

STDMETHODIMP CLuaFilter::GetDescriptionText( long /*index*/, BSTR* /*description*/ )
{
	return E_NOTIMPL;
}

STDMETHODIMP CLuaFilter::GetInitialExtent( long* line, long* startCol, long* endCol )
{
	if( !line || !startCol || !endCol )
        return E_POINTER;

	// initialise to sensible defaults
    *line = *startCol = *endCol = 0;

	// get the current line where the cursor is
    HRESULT hr = m_view->GetCaretPos( line, startCol );
    if( SUCCEEDED( hr ) )
        *endCol = *startCol;

	// done
    return hr;
}

STDMETHODIMP CLuaFilter::GetBestMatch( WCHAR const* /*soFar*/, long /*length*/, long* /*index*/, DWORD* /*flags*/ )
{
	return E_NOTIMPL;
}

STDMETHODIMP CLuaFilter::OnCommit( WCHAR const* /*soFar*/, long /*index*/, BOOL /*selected*/, WCHAR /*commit*/, BSTR* /*completeWord*/ )
{
	return E_NOTIMPL;
}

STDMETHODIMP_( void ) CLuaFilter::Dismiss()
{
}

void CLuaFilter::ResetList()
{
	// clear the list
	std::set<std::wstring, StringSorter> idents;

	// add the lua keywords
	idents.insert( L"and" );
	idents.insert( L"break" );
	idents.insert( L"do" );
	idents.insert( L"else" );
	idents.insert( L"elseif" );
	idents.insert( L"end" );
	idents.insert( L"false" );
	idents.insert( L"for" );
	idents.insert( L"function" );
	idents.insert( L"if" );
	idents.insert( L"in" );
	idents.insert( L"local" );
	idents.insert( L"nil" );
	idents.insert( L"not" );
	idents.insert( L"or" );
	idents.insert( L"repeat" );
	idents.insert( L"return" );
	idents.insert( L"then" );
	idents.insert( L"true" );
	idents.insert( L"until" );
	idents.insert( L"while" );

	// scan each line of text storing identifiers as they're found
	CComPtr<IVsTextLines> lines;
	HRESULT hr = m_view->GetBuffer( &lines );
	if( SUCCEEDED( hr ) )
	{
		long currentLine = 0, currentColumn = 0;
	    m_view->GetCaretPos( &currentLine, &currentColumn );

		long lineCount = 0;
		lines->GetLineCount( &lineCount );

		LINEDATA data;
		for( long line = 0; line < lineCount; ++line )
		{
			// get the data for a line
			hr = lines->GetLineData( line, &data, 0 );
			if( line != currentLine && SUCCEEDED( hr ) )
			{
				// scan it for keywords
				std::wstring current;
				for( long start = 0; start < data.iLength; )
				{
					wchar_t c = data.pszText[start];

					if( ( 'a' <= c && c <= 'z' )
						|| ( 'A' <= c && c <= 'Z' )
						|| c == '_' )
					{
						long end = start + 1;
						for( ; end < data.iLength; ++end )
						{
							c = data.pszText[end];
							if( !( 'a' <= c && c <= 'z' )
								&& !( 'A' <= c && c <= 'Z' )
								&& !( '0' <= c && c <= '9' )
								&& c != '_' 
							)
								break;
						}

						idents.insert( std::wstring( data.pszText + start, data.pszText + end ) );

						start = end;
					}
					else
						++start;
				}
			}
		}
	}

	// make an array-based copy of the sorted list
	m_identifiers.clear();
	std::copy( idents.begin(), idents.end(), std::back_inserter( m_identifiers ) );
}
