// LuaLanguageInfo.cpp : Implementation of CLuaLanguageInfo

#include "stdafx.h"
#include "LuaLanguageInfo.h"
#include "LuaColouriser.h"
#include "LuaWindowManager.h"


// The language info service ID: {04DA40AA-AE4B-49B4-8033-CF3D0BCEE2EC}
GUID const SID_LuaLanguageInfo = { 0x04DA40AA,0xAE4B,0x49B4,{0x80,0x33,0xCF,0x3D,0x0B,0xCE,0xE2,0xEC} };


// CLuaLanguageInfo

STDMETHODIMP CLuaLanguageInfo::GetLanguageName( BSTR *name )
{
	// check argument
	if( !name )
		return E_INVALIDARG;

	// allocate the string
	*name = SysAllocString( L"Lua" );
	return *name ? S_OK : E_UNEXPECTED;
}

STDMETHODIMP CLuaLanguageInfo::GetFileExtensions( BSTR *extensions )
{
	// check argument
	if( !extensions )
		return E_INVALIDARG;

	// allocate the string
	*extensions = SysAllocString( L".lua;" );
	return *extensions ? S_OK : E_UNEXPECTED;
}

STDMETHODIMP CLuaLanguageInfo::GetColorizer( IVsTextLines *buffer, IVsColorizer **colouriser )
{
	// check arguments
	if( !buffer || !colouriser )
		return E_INVALIDARG;

	// create a colouriser
	CComPtr<IVsColorizer> temp;
	HRESULT hr = CLuaColouriser::CreateInstance( &temp );
	if( FAILED( hr ) )
		return hr;

	// assign the buffer
	CLuaColouriser* parent = static_cast<CLuaColouriser*>( temp.p );
	parent->SetTextLines( buffer );

	// set the argument
	*colouriser = temp.Detach();

	// done ok
	return S_OK;
}

STDMETHODIMP CLuaLanguageInfo::GetCodeWindowManager( IVsCodeWindow* window, IVsCodeWindowManager** manager )
{
	// check arguments
	if( !manager )
		return E_INVALIDARG;

	// create a window manager
	CComPtr<IVsCodeWindowManager> temp;
	HRESULT hr = CLuaWindowManager::CreateInstance( &temp );
	if( FAILED( hr ) )
		return hr;

	// assign the view
	CLuaWindowManager* parent = static_cast<CLuaWindowManager*>( temp.p );
	parent->SetWindow( window );

	// set the argument
	*manager = temp.Detach();

	// done ok
	return S_OK;
}

STDMETHODIMP CLuaLanguageInfo::GetItemCount( int* count )
{
	if( !count )
		return E_POINTER;

	*count = int( m_colourables.size() );
	return S_OK;
}

STDMETHODIMP CLuaLanguageInfo::GetColorableItem( int index, IVsColorableItem **item )
{
	if( index < 1 || int( m_colourables.size() ) < index )
		return E_INVALIDARG;
	if( !item )
		return E_POINTER;

	*item = m_colourables[index - 1].m_T;
	( *item )->AddRef();
	return S_OK;
}

struct LUA_ITEM_DEF
{
	COLORINDEX foreground;
	COLORINDEX background;
	DWORD fontFlags;
	wchar_t const* name;
};

HRESULT CLuaLanguageInfo::CreateColourables()
{
	// set up some definitions (for the LuaScannerType enumeration)
	LUA_ITEM_DEF defs[] = 
	{
		{ CI_USERTEXT_FG,	CI_USERTEXT_BK,	FF_DEFAULT, L"Text" }, 
		{ CI_USERTEXT_FG,	CI_USERTEXT_BK,	FF_DEFAULT, L"Identifier" }, 
		{ CI_DARKGREEN,		CI_USERTEXT_BK,	FF_DEFAULT, L"Comment" }, 
		{ CI_BLUE,			CI_USERTEXT_BK,	FF_DEFAULT, L"Keyword" }, 
		{ CI_USERTEXT_FG,	CI_USERTEXT_BK,	FF_DEFAULT, L"Number" }, 
		{ CI_USERTEXT_FG,	CI_USERTEXT_BK,	FF_DEFAULT, L"String" }, 
		{ CI_USERTEXT_FG,	CI_USERTEXT_BK,	FF_DEFAULT, L"Operator" }, 
		{ CI_USERTEXT_FG,	CI_USERTEXT_BK, 0 ,0 }
	};

	// create each item
	for( int i = 0; defs[i].name; ++i )
	{
		CComPtr<IVsColorableItem> item;
		HRESULT hr = CLuaColourableItem::CreateInstance( &item );
		if( FAILED( hr ) )
			return hr;

		CLuaColourableItem* parent = static_cast<CLuaColourableItem*>( item.p );
		parent->Create(
			defs[i].foreground, 
			defs[i].background, 
			defs[i].fontFlags, 
			defs[i].name
		);

		m_colourables.push_back( item );
	}

	// done!
	return S_OK;
}

void CLuaColourableItem::Create( COLORINDEX foreground, COLORINDEX background, DWORD fontFlags, std::wstring const& name )
{
	m_foreground = foreground;
	m_background = background;
	m_fontFlags = fontFlags;
	m_name = name;
}

STDMETHODIMP CLuaColourableItem::GetDefaultColors( COLORINDEX* foreground, COLORINDEX* background )
{
	if( !foreground || !background )
		return E_POINTER;

	*foreground = m_foreground;
	*background = m_background;
	return S_OK;
}

STDMETHODIMP CLuaColourableItem::GetDefaultFontFlags( DWORD* fontFlags )
{
	if( !fontFlags )
		return E_POINTER;

	*fontFlags = m_fontFlags;
	return S_OK;
}

STDMETHODIMP CLuaColourableItem::GetDisplayName( BSTR *name )
{
	if( !name )
		return E_POINTER;

	*name = SysAllocString( m_name.c_str() );
	return S_OK;
}
