#include "stdafx.h"
#include "LuaScanner.h"

LuaScanner::LuaScanner()
  : m_buffer( 0 ), 
    m_length( 0 ), 
	m_state( STATE_DEFAULT ), 
	m_level( 0 ), 
	m_pos( 0 ), 
	m_typesLength( 0 )
{
	// store all the lua keywords
	m_keywords.insert( L"and" );
	m_keywords.insert( L"break" );
	m_keywords.insert( L"do" );
	m_keywords.insert( L"else" );
	m_keywords.insert( L"elseif" );
	m_keywords.insert( L"end" );
	m_keywords.insert( L"false" );
	m_keywords.insert( L"for" );
	m_keywords.insert( L"function" );
	m_keywords.insert( L"if" );
	m_keywords.insert( L"in" );
	m_keywords.insert( L"local" );
	m_keywords.insert( L"nil" );
	m_keywords.insert( L"not" );
	m_keywords.insert( L"or" );
	m_keywords.insert( L"repeat" );
	m_keywords.insert( L"return" );
	m_keywords.insert( L"then" );
	m_keywords.insert( L"true" );
	m_keywords.insert( L"until" );
	m_keywords.insert( L"while" );

	// store all the lua operators
    m_operators.insert( L"+" );
	m_operators.insert( L"-" );
	m_operators.insert( L"*" );
	m_operators.insert( L"/" );
	m_operators.insert( L"^" );
	m_operators.insert( L"=" );
	m_operators.insert( L"==" );
	m_operators.insert( L"~=" );
	m_operators.insert( L"<=" );
	m_operators.insert( L">=" );
	m_operators.insert( L"<" );
	m_operators.insert( L">" );
	m_operators.insert( L"(" );
	m_operators.insert( L")" );
	m_operators.insert( L"{" );
	m_operators.insert( L"}" );
	m_operators.insert( L"[" );
	m_operators.insert( L"]" );
	m_operators.insert( L";" );
	m_operators.insert( L":" );
	m_operators.insert( L"," );
	m_operators.insert( L"." );
	m_operators.insert( L".." );
	m_operators.insert( L"..." );
	m_maxOperatorLength = 3;
}

int LuaScanner::ScanBuffer( wchar_t const* buffer, int length, int state )
{
	// store the buffer
	m_buffer = buffer;
	m_length = length;
	m_pos = 0;

	// check for storage for the types array
	int const typesMax = m_length + 1;
	if( m_typesLength < typesMax )
	{
		m_types.reset( new int[typesMax] );
		m_typesLength = typesMax;
	}

	// set up the state
	m_state = static_cast<STATE_TYPE>( state & 0xff );
	m_level = state >> 8;

	// scan 
	Scan();

	// append the final type
	switch( m_state )
	{
	case STATE_COMMENT_BLOCK:
		m_types[m_length] = LUA_COMMENT;
		break;

	default: // STATE_DEFAULT
		m_types[m_length] = LUA_TEXT;
		break;
	}

	// return the end-of-line state
	return ( m_level << 8 ) | m_state;
}

void LuaScanner::Scan()
{
	while( IsValid() )
	{
		switch( m_state )
		{
		case STATE_COMMENT_BLOCK: 
			ScanCommentBlock();
			break;

		case STATE_STRING_SQ: 
			ScanSingleQuoteString();
            break;

		case STATE_STRING_DQ:
			ScanDoubleQuoteString();
            break;

		case STATE_STRING_BK:
			ScanBracketString();
            break;

		default: // STATE_DEFAULT
			ScanText();
			break;
		}
	}
}

void LuaScanner::ScanText()
{
	if( IsLetter() )
	{
		// extract an identifier
		ExtractIdentifier();
	}
	else if( IsDigit() || IsSignedDigit() )
	{
		// extract a number
		ExtractNumber();
	}
	else if( IsCommentBlock() )
	{
		// enter a comment block
		m_state = STATE_COMMENT_BLOCK;
		m_level = 1;

		// advance
		Advance( LUA_COMMENT, 4 );
	}
	else if( IsComment() )
	{
		// extract the comment
		ExtractComment();
	}
	else if( IsSingleQuote() )
	{
		// enter a string block
		m_state = STATE_STRING_SQ;

		// advance
		Advance( LUA_STRING );
	}
	else if( IsDoubleQuote() )
	{
		// enter a string block
		m_state = STATE_STRING_DQ;

		// advance
		Advance( LUA_STRING );
	}
	else if( IsDoubleBracketStart() )
	{
		// enter a string block
		m_state = STATE_STRING_BK;
		m_level = 1;

		// advance
		Advance( LUA_STRING, 2 );
	}
	else
	{
		// otherwise plain text or an operator
		ExtractText();
	}
}

void LuaScanner::ScanCommentBlock()
{
	if( IsDoubleBracketStart() )
	{
		// nested comment evilness
		++m_level;

		// advance
		Advance( LUA_COMMENT, 2 );
	}
	else if( IsDoubleBracketEnd() )
	{
		// reduce the nested level
		if( --m_level == 0 )
			m_state = STATE_DEFAULT;

		// advance
		Advance( LUA_COMMENT, 2 );
	}
	else
	{
		// otherwise comment
		Advance( LUA_COMMENT );
	}
}

void LuaScanner::ScanDoubleQuoteString()
{
	if( IsEscapedEscape() || IsEscapedDoubleQuote() )
	{
		// ignore it
		Advance( LUA_STRING, 2 );
	}
	else if( IsDoubleQuote() )
	{
		// end the string
		m_state = STATE_DEFAULT;

		// advance
		Advance( LUA_STRING );
	}
	else
	{
		// otherwise advance
		Advance( LUA_STRING );
	}
}

void LuaScanner::ScanSingleQuoteString()
{
	if( IsEscapedEscape() || IsEscapedSingleQuote() )
	{
		// ignore it
		Advance( LUA_STRING, 2 );
	}
	else if( IsSingleQuote() )
	{
		// end the string
		m_state = STATE_DEFAULT;

		// advance
		Advance( LUA_STRING );
	}
	else
	{
		// otherwise advance
		Advance( LUA_STRING );
	}
}

void LuaScanner::ScanBracketString()
{
	if( IsDoubleBracketStart() )
	{
		// nested string evilness
		++m_level;

		// advance
		Advance( LUA_STRING, 2 );
	}
	else if( IsDoubleBracketEnd() )
	{
		// pop a level
		if( --m_level == 0 )
			m_state = STATE_DEFAULT;

		// advance
		Advance( LUA_STRING, 2 );
	}
	else
	{
		// string data
		Advance( LUA_STRING );
	}
}

void LuaScanner::ExtractComment()
{
	// early-out to the end of the line
	while( IsValid() )
		m_types[m_pos++] = LUA_COMMENT;
}

void LuaScanner::ExtractIdentifier()
{
	// store the start
	int start = m_pos;

	// extract the identifier
	while( IsValid() && 
		( IsLetter() || IsDigit() ) 
	)
	{
		Advance( LUA_IDENTIFIER );
	}

	// check for a keyword
	std::wstring keyword( m_buffer + start, m_buffer + m_pos );
	if( m_buffer[start] == '_' || m_keywords.find( keyword ) != m_keywords.end() )
	{
		for( int i = start; i < m_pos; ++i )
			m_types[i] = LUA_KEYWORD;
	}
}

void LuaScanner::ExtractNumber()
{
	// mantissa = \-?{digit}+(\.{digit}*)?
	if( m_buffer[m_pos] == '-' )
		Advance( LUA_NUMBER );
	while( IsValid() && IsDigit() )
		Advance( LUA_NUMBER );
	if( IsValid() && m_buffer[m_pos] == '.' )
		Advance( LUA_NUMBER );
	while( IsValid() && IsDigit() )
		Advance( LUA_NUMBER );

	// exponent = [eE]{sign}?{digit}+
	if( IsValid() && ( m_buffer[m_pos] == 'e' || m_buffer[m_pos] == 'E' ) )
	{
		Advance( LUA_NUMBER );
		if( IsValid() && m_buffer[m_pos] == '-' )
			Advance( LUA_NUMBER );
		while( IsValid() && IsDigit() )
			Advance( LUA_NUMBER );
	}
}

void LuaScanner::ExtractText()
{
	// check for an operator
	int maxlen = std::min( m_maxOperatorLength, m_length - m_pos );
	for( int len = maxlen; len > 0; --len )
	{
		std::wstring candidate( m_buffer + m_pos, m_buffer + m_pos + len );
		if( m_operators.find( candidate ) != m_operators.end() )
		{
			Advance( LUA_OPERATOR, len );
			return;
		}
	}

	// must be plain text
	Advance( LUA_TEXT );
}

bool LuaScanner::IsValid() const
{
	return m_pos < m_length;
}

bool LuaScanner::IsLetter() const
{
	return ( 'a' <= m_buffer[m_pos] && m_buffer[m_pos] <= 'z' )
		|| ( 'A' <= m_buffer[m_pos] && m_buffer[m_pos] <= 'Z' )
		|| m_buffer[m_pos] == '_'
		;
}

bool LuaScanner::IsDigit() const
{
	return ( '0' <= m_buffer[m_pos] && m_buffer[m_pos] <= '9' );
}

bool LuaScanner::IsSignedDigit() const
{
	if( m_length - m_pos < 2 )
		return false;

	return m_buffer[m_pos] == '-'
		&& '0' <= m_buffer[m_pos + 1] && m_buffer[m_pos + 1] <= '9';
}

bool LuaScanner::IsCommentBlock() const
{
	return Matches( L"--[[" );
}

bool LuaScanner::IsDoubleBracketStart() const 
{
	return Matches( L"[[" );
}

bool LuaScanner::IsDoubleBracketEnd() const 
{
	return Matches( L"]]" );
}

bool LuaScanner::IsComment() const
{
	return Matches( L"--" );
}

bool LuaScanner::IsSingleQuote() const
{
	return m_buffer[m_pos] == '\'';
}

bool LuaScanner::IsDoubleQuote() const
{
	return m_buffer[m_pos] == '\"';
}

bool LuaScanner::IsEscapedSingleQuote() const 
{
	return Matches( L"\\'" );
}

bool LuaScanner::IsEscapedDoubleQuote() const
{
	return Matches( L"\\\"" );
}

bool LuaScanner::IsEscapedEscape() const
{
	return Matches( L"\\\\" );
}

bool LuaScanner::Matches( wchar_t const* str ) const
{
	// check for space
	size_t length = wcslen( str );
	if( size_t( m_length - m_pos ) < length )
		return false;

	// check substring
	return ( wcsncmp( str, m_buffer + m_pos, length ) == 0 );
}

void LuaScanner::Advance( int type, int count )
{
	for( int i = 0; i < count; ++i )
		m_types[m_pos++] = type;
}
